# -*- coding: utf-8 -*-

## FL.py provides support for "at Function Level" style of Python programming.
## Copyright (C) 2009 Alberto Paoluzzi, Michele Vicentino

## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.

## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## Lesser General Public License for more details.

## You should have received a copy of the GNU Lesser General Public
## License along with this library; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
This module contains basic FL and PLaSM combinators and functions
providing some backward compatibility with previous PLaSM versions. 
"""

import math
import types

def __evalprint__(string):
    print string + " => ", eval(string)

## --------------------------------------------------
## --Define CONStants--------------------------------
## --------------------------------------------------

PI = math.pi
SIN = math.sin
SINH = math.sinh
ASIN = math.asin
COS = math.cos
COSH = math.cosh
ACOS = math.acos
TAN = math.tan
TANH = math.tanh
ATAN = math.atan
SQRT = math.sqrt
EXP = math.exp
LN = math.log
CEIL = math.ceil
FLOOR = math.floor
ABS = abs
CHAR = chr
ORD = ord

if __name__ == "__main__":
    print "\n## --Miscellanea functions----------------------------"
    __evalprint__(""" ORD(CHAR(87)) == 87 """)
    

## --------------------------------------------------
## --FL IDentity Function----------------------------
## --------------------------------------------------

def ID (anyValue):
    """
    ID: Standard core of PyPLaSM
    returns the argument unchanged
    (a::tt) -> (tt)
    """
    return anyValue

if __name__ == "__main__":
    print "\n## --FL IDentity Function----------------------------"
    __evalprint__(""" ID(5) """)
    __evalprint__(""" ID("string")  """)
    __evalprint__(""" ID(lambda x: x*x)  """)
    __evalprint__(""" ID(lambda x: x*x)(3) """)

## --------------------------------------------------
## --FL CONStant Function----------------------------
## --------------------------------------------------

def K(a):
    """
    K:  Standard core of PyPLaSM
    constant functional. Always returns the first argument, for any v alue of the second
    (a::tt)(b::tt) -> tt
    """
    return (lambda b: a)

if __name__ == "__main__":
    print "\n## --FL CONStant Function----------------------------"
    __evalprint__(""" K(5) """)
    __evalprint__(""" K(5)(100) """)
    __evalprint__(""" K(5)(lambda x: x*x) """)
    __evalprint__(""" K(lambda x: x*x)(100) """)
    __evalprint__(""" K(lambda x: x*x)(100)(100) """)

## --------------------------------------------------
## --FL DISTRibute Left------------------------------
## --------------------------------------------------

def DISTL(Pair):
    """
    DISTL:  Standard core of PyPLaSM
    distribute left. Returns the pair sequence with x} and the elements of seq
    (x::tt,seq::isseq) -> (isseqof:ispair)
    """
    Element, Indexable = Pair[0], Pair[1]
    return [ [Element, e] for e in Indexable ]

if __name__ == "__main__":
    print "\n## --FL DISTRibute Left------------------------------"
    __evalprint__(""" DISTL([10,[1,2,3]]) """)
    __evalprint__(""" DISTL([10,[]]) """)
    __evalprint__(""" DISTL((10,())) """)
    __evalprint__(""" DISTL([10,range(10)]) """)
    __evalprint__(""" DISTL([SIN,range(5)]) """)

## --------------------------------------------------
## --FL DISTRibute Right-----------------------------
## --------------------------------------------------

def DISTR(Pair):
    """
    DISTR:  Standard core of PyPLaSM
    distribute right. Returns the pair sequence with the elements of seq} and x}
    (seq::isseq,x::tt) -> (isseqof:ispair)
    """
    Indexable, Element = Pair[0], Pair[1]
    return [ [e, Element] for e in Indexable ]

if __name__ == "__main__":
    print "\n## --FL DISTRibute Right-----------------------------"
    __evalprint__(""" DISTR([[1,2,3],10]) """)
    __evalprint__(""" DISTR([[],10]) """)
    __evalprint__(""" DISTR([range(10),10]) """)
    __evalprint__(""" DISTR([range(5),SIN]) """)
    __evalprint__(""" DISTR([(1,2,'a','b'),100]) """)

## --------------------------------------------------
## --FL COMPosition----------------------------------
## --------------------------------------------------

def COMP(args):
    """
    COMP:   Standard core of PyPLaSM
    composition. Returns the composition of the functions in the argument sequence
    (funs::isseqof:isfun) -> (isfun)
    """
    if (len(args) > 1):
        return (lambda x: COMP(args[0:-1])(args[-1](x)))
    elif (len(args) == 1): 
        return (lambda x: args[0](x))
    else:
        return (lambda x: x)
    
if __name__ == "__main__":
    print "\n## --FL COMPosition----------------------------------"
    __evalprint__(""" COMP([ID,ID,ID]) """)
    __evalprint__(""" COMP([ID,ID,SIN])(0.0) """)
    __evalprint__(""" COMP([ID,COS,SIN])(1.0) """)
    __evalprint__(""" COMP([K,ID,SIN])(1.0)  """)
    __evalprint__(""" COMP([K,ID,SIN])(PI/2)(0.0)  """)
    __evalprint__(""" COMP([SIN,SIN,ID])(0.0) """)

## --------------------------------------------------
## --FL Apply-to-all---------------------------------
## --------------------------------------------------

def AA (f):
    """
    AA: Standard core of PyPLaSM
    applies fun to each element of the args sequence
    (fun::isfun)(args::isseq) -> (isseq)
    """
    return lambda List: map(f, List)

if __name__ == "__main__":
    print "\n## --FL Apply-to-all---------------------------------"
    __evalprint__(""" AA(lambda x: x*x)(range(10)) """)
    __evalprint__(""" AA(COMP([SQRT,lambda x: x*x]))(range(10)) """)
#    __evalprint__(""" AA(COMP([SQRT,SIN]))(range(10)) """)

## --------------------------------------------------
## --FL Apply----------------------------------------
## --------------------------------------------------

def APPLY(args):
    """
    APPLY:  Standard core of PyPLaSM
    returns the result of the expression fun:value
    (fun::isfun,value::tt) -> (tt)
    """
    f,x = args
    return apply(f,[x])

if __name__ == "__main__":
    print "\n## --FL Apply----------------------------------------"
    __evalprint__(""" APPLY([SQRT,81]) """)
    __evalprint__(""" AA(APPLY)([[SQRT,9],[SQRT,16],[SQRT,25]]) """)
#?    __evalprint__(""" COMP([AA(APPLY),D1(SQRT)])(range(10)) """)

## --------------------------------------------------
## --n-ary addition----------------------------------
## --------------------------------------------------

def SUM(args):
    return reduce(lambda x,y: x+y, args)

if __name__ == "__main__":
    print "\n## --n-ary addition----------------------------------"
    __evalprint__(""" SUM([1,2,3,4,5]) """)
    __evalprint__(""" SUM(range(11)) """)

## --------------------------------------------------
## --n-ary DIFFerence--------------------------------
## --------------------------------------------------

def DIFF(args):
    return reduce(lambda x,y: x-y, args)

if __name__ == "__main__":
    print "\n## --n-ary DIFFerence--------------------------------"
    __evalprint__(""" DIFF([1,2,3,4,5]) """)
    __evalprint__(""" DIFF(range(11)) """)

## --------------------------------------------------
## --n-ary PRODuct-----------------------------------
## --------------------------------------------------

def PROD(args):
    return reduce(lambda x,y: x*y, args)

if __name__ == "__main__":
    print "\n## --n-ary PRODuct-----------------------------------"
    __evalprint__(""" PROD([1,2,3,4,5]) """)
    __evalprint__(""" PROD(range(11)) """)
    __evalprint__(""" PROD([1,2,3,4]) """)

## --------------------------------------------------
## --n-ary DIVision----------------------------------
## --------------------------------------------------

def DIV(args):
    return reduce(lambda x,y: x/y, args)

if __name__ == "__main__":
    print "\n## --n-ary DIVision----------------------------------"
    __evalprint__(""" DIV([1.0,2.0,3.0]) """)


## --------------------------------------------------
## --POW function----------------------------------
## --------------------------------------------------

def POW(args):
    x,y = args
    return pow(x,y)

if __name__ == "__main__":
    print "\n## --Power elevation----------------------------------"
    __evalprint__(""" POW([2,3]) """)
    __evalprint__(""" POW([2.0,3.0]) """)


## --------------------------------------------------
## --FL Sequences------------------------------------
## --------------------------------------------------

def REVERSE(List):
    """
    REVERSE:    Standard core of PyPLaSM
    returns a sequence in reverse order
    (seq::isseq) -> (isseq)
    """
    if (len(List) > 1): return [List[-1]] + REVERSE(List[:-1])
    else: return List

if __name__ == "__main__":
    print "\n## --FL Sequences: REVERSE-----------------------------"
    __evalprint__(""" REVERSE([10,20,30]) """)
    __evalprint__(""" REVERSE([10]) """)
    __evalprint__(""" REVERSE([]) """)

LEN = len


def CAT (List):
    """
    CAT:    Standard core of PyPLaSM
    catenates a sequence of sequences, by eliminating a level of angled parenthesis
    (seqs::isseqof:isseq) -> (isseq)
    """
    return SUM (List)

if __name__ == "__main__":
    print "\n## --FL Sequences: CAT-----------------------------"
    __evalprint__(""" CAT([[10,20,30],[10,20,30],[10,20,30]]) """)
    __evalprint__(""" LEN(CAT([[10,20,30],[10,20,30],[10,20,30]])) """)

def TRANS (List):
    """
    TRANS:  Standard core of PyPLaSM
    transpose a sequence of sequences of the same length. The elements may be of arbitrary type
    (seq::ismat) -> (ismat)
    """
    return map(list, zip(*List))

if __name__ == "__main__":
    print "\n## --FL Sequences: TRANS-----------------------------"
    __evalprint__(""" TRANS([[10,20,30],[10,20,30],[10,20,30]]) """)

def FIRST (List):
    """
    FIRST:  Standard core of PyPLaSM
    returns the first element of the sequence given as argument.
    (seq::and ~ [isseq,not ~ isnull]) -> (tt)
    """
    return List[0]

if __name__ == "__main__":
    print "\n## --FL Sequences: FIRST-----------------------------"
    __evalprint__(""" FIRST([10,20,30]) """)

def LAST (List):
    """
    LAST:   Standard core of PyPLaSM
    returns the last element of the nonempty sequence argument
    (seq::and ~ [isseq,not ~ null]) -> (tt)
    """
    return List[-1]

if __name__ == "__main__":
    print "\n## --FL Sequences: LAST-----------------------------"
    __evalprint__(""" LAST([10,20,30]) """)

def TAIL (List):
    """
    TAIL:   Standard core of PyPLaSM
    returns the non-empty argument sequence but its first element
    (seq::and ~ [isseq,not ~ isnull]) -> (isseq)
    """
    return List[1:]

if __name__ == "__main__":
    print "\n## --FL Sequences: TAIL-----------------------------"
    __evalprint__(""" TAIL([10,20,30]) """)

def RTAIL (List):
    """
    TAIL:   Standard core of PyPLaSM
    returns the non-empty argument sequence but its last element
    (seq::and ~ [isseq,not ~ isnull]) -> (isseq)
    """
    return List[:-1]

if __name__ == "__main__":
    print "\n## --FL Sequences: RTAIL-----------------------------"
    __evalprint__(""" RTAIL([10,20,30]) """)

def AR (Pair):
    """
    AR: Standard core of PyPLaSM
    append right. appends elem} on the right of seq}
    (seq::isseq) -> (elem::tt) -> (isseq)
    """
    seq,elem = Pair[0],Pair[1]
    return seq + [elem]
     
if __name__ == "__main__":
    print "\n## --FL Sequences: REVERT-----------------------------"
    __evalprint__(""" AR([[10,20,30],400]) """)
    __evalprint__(""" AR([[],400]) """)

def AL (Pair):
    """
    AL: Standard core of PyPLaSM
    append left. appends elem} on the left of seq}
    (elem::tt) -> (seq::isseq) -> (isseq)
    """
    elem,seq = Pair[0],Pair[1]
    return [elem] + seq
    
if __name__ == "__main__":
    print "\n## --FL Sequences: REVERT-----------------------------"
    __evalprint__(""" AL([400,[10,20,30]]) """)
    __evalprint__(""" AL([400,[]]) """)

def LIST (x):
    """
    LIST:   Standard core of PyPLaSM
    returns the sequence with only the argument a}. Alias for [id]}
    (a::tt) -> (isseq)
    """
    return [x]

if __name__ == "__main__":
    print "\n## --FL Sequences: LIST------------------------------"
    __evalprint__(""" LIST(100) """)
    __evalprint__(""" LIST(ID) """)
    __evalprint__(""" COMP([LIST,ID,SIN])(PI/2) """)

## --------------------------------------------------
## --FL Insert Right/Left----------------------------
## --------------------------------------------------

def greater (x,y):
    if x > y: return x
    else: return y

def I2(f):
    """
    I2: Standard core of PyPLaSM
    insert right. Recursive functional allowing to apply a binary operator f to n arguments
    (f::isfun)(args::and ~ [isseq,not~isnull]) -> tt
    """
    return lambda args: reduce(f, args)
    
if __name__ == "__main__":
    print "\n## --FL Insert Right/Left----------------------------"
    __evalprint__(""" I2(SUM) """)
    __evalprint__(""" I2(greater)([2,4,6,8,10]) """)

def I1(f): 
    """
    I1: Standard core of PyPLaSM
    insert left. Recursive functional allowing to apply a binary operator f to n arguments
    (f::isfun)(args::and ~ [isseq,not~isnull]) -> tt
    """
    return COMP([I2(f),REVERSE])

if __name__ == "__main__":
    print "\n## --FL Insert Right/Left----------------------------"
    __evalprint__(""" I1(greater) """)
    __evalprint__(""" I1(greater)([2,400,6,8,10]) """)

def BIGGEST (List): return I2(greater)(List)

if __name__ == "__main__":
    print "\n## --FL Insert Right/Left----------------------------"
    __evalprint__(""" BIGGEST([2,400,6,8,10]) """)
    __evalprint__(""" BIGGEST([2,400]) """)
    __evalprint__(""" BIGGEST([2,-400]) """)

## --------------------------------------------------
## --FL CONStruction---------------------------------
## --------------------------------------------------

def CONS(args):
    """
    CONS:    Standard core of PyPLaSM
    construction. Applies a function sequence to an argument.
    Retun the sequence of applications
    (funs::isseqof:isfun)(x::tt) -> (isseq)
    """
    return lambda x: AA(APPLY)(DISTR([args,x]))

if __name__ == "__main__":
    print "\n## --FL CONStruction---------------------------------"
    __evalprint__(""" CONS([SUM,DIFF,PROD,DIV]) """)
    __evalprint__(""" CONS([SUM,DIFF,PROD,DIV])([5,10]) """)
    __evalprint__(""" CONS([SUM,DIFF,PROD,DIV])([5.0,10.0]) """)
    __evalprint__(""" CONS([SIN,COS,TAN])(ACOS(-1)/4.0) """)
    __evalprint__(""" CONS([SIN,COS,TAN])(PI/4) """)

## --------------------------------------------------
## --PLASM range builders----------------------------
## --------------------------------------------------

def INTSTO (n): 
    """
    INTSTO: Standard core of PyPLaSM
    integers to. return the integer sequence to n, or the sequence -1 .. n if n<0, or the empty sequence if n=0
    (n::isint) -> isseqof:isint
    """
    return AA(lambda x: x+1)(range(n))
    
    print "\n## --PLASM range builders----------------------------"
    __evalprint__(""" INTSTO(10) """)
    
def FROMTO (args):
    """
    FROMTO: Standard core of PyPLaSM
    returns the integer sequence from m to n. 
    When m < n  the returned sequence is empty.
    (m,n::isint) -> isseqof:isint
    """
    n,m = args
    return range(n,m+1)

if __name__ == "__main__":
    print "\n## --PLASM range builders----------------------------"
    __evalprint__(""" FROMTO([5,15]) """)
    __evalprint__(""" FROMTO([-5,5]) """)

## --------------------------------------------------
## --PLASM  selectors -------------------------------
## --------------------------------------------------

def SEL (n): 
    """
    SEL:    Standard core of PyPLaSM
    returns the n-th element of a sequence. An exception is raised if n > len(seq)
    (i::isintpos)(seq::isseq) -> tt
    """
    return lambda List: List[n-1]

S1 = SEL(1)
S2 = SEL(2)
S3 = SEL(3)
S4 = SEL(4)
S5 = SEL(5)
S6 = SEL(6)
S7 = SEL(7)
S8 = SEL(8)
S9 = SEL(9)
S10 = SEL(10)

if __name__ == "__main__":
    print "\n## --PLASM  selectors -------------------------------"
    __evalprint__(""" SEL(10)(range(101,200)) """)
    __evalprint__(""" S1(range(101,200)) """)
    __evalprint__(""" S2(range(101,200)) """)
    __evalprint__(""" S3(range(101,200)) """)
    __evalprint__(""" S4(range(101,200)) """)
    __evalprint__(""" S5(range(101,200)) """)
    __evalprint__(""" S6(range(101,200)) """)
    __evalprint__(""" S7(range(101,200)) """)
    __evalprint__(""" S8(range(101,200)) """)
    __evalprint__(""" S9(range(101,200)) """)
    __evalprint__(""" S10(range(101,200)) """)

## --------------------------------------------------
## --PLASM  repeat operators-------------------------
## --------------------------------------------------

def N (n): 
    """
    N: Standard core of PyPLaSM
    repetition operator. Returns the sequence with n repetitions of arg
    (n::isintpos)(arg::tt) -> (isseq)
    """
    return lambda List: [List]*n

if __name__ == "__main__":
    print "\n## --PLASM  repeat operators: N----------------------"
    __evalprint__(""" N(5)([1,2,3]) """)
    __evalprint__(""" N(0)([1,2,3]) """)

def NN (n): 
    """
    NN:   Standard core of PyPLaSM
    sequence repetition operator. Returns the sequence CAT(N(seq))
    (n::isintpos)(seq::tt) -> (isseq)
    """
    return lambda List: List*n

if __name__ == "__main__":
    print "\n## --PLASM  repeat operators: NN----------------------"
    __evalprint__(""" NN(5)([1,2,3]) """)
    __evalprint__(""" NN(0)([1,2,3]) """)

## --------------------------------------------------
## --PLASM  logical operators------------------------
## --------------------------------------------------

def And(x,y): return x and y
def Or(x,y): return x or y
def Not(x): return not x


AND = I2(And)
OR = I2(Or)
NOT = AA(Not)
FALSE = False
TRUE = True
TT = K(TRUE)

if __name__ == "__main__":
    print "\n## --PLASM  logical operators------------------------"
    __evalprint__(""" AND """)
    __evalprint__(""" AND([True]) """)
    __evalprint__(""" AND([False]) """)
    __evalprint__(""" AND([True,True]) """)
    __evalprint__(""" AND([True,True,True,True,True]) """)
    __evalprint__(""" AND([True,True,False,True,True]) """)
    __evalprint__(""" COMP([Not,AND]) """)
    __evalprint__(""" COMP([Not,AND])([True,True]) """)
    __evalprint__(""" COMP([Not,AND])([True,True,True,True,True]) """)
    __evalprint__(""" COMP([Not,AND])([True,True,False,True,True]) """)
    __evalprint__(""" OR """)
    __evalprint__(""" OR([True,True]) """)
    __evalprint__(""" OR([False,False]) """)
    __evalprint__(""" OR([False,False,False,False,False]) """)
    __evalprint__(""" OR([False,True,False,False,False]) """)
    __evalprint__(""" OR([True,True,False,True,True]) """)
    __evalprint__(""" NOT """)
    __evalprint__(""" NOT([True,True]) """)
    __evalprint__(""" NOT([False,False]) """)
    __evalprint__(""" NOT([False,False,False,False,False]) """)
    __evalprint__(""" NOT([False,True,False,False,False]) """)
    __evalprint__(""" NOT([True,True,False,True,True]) """)

## --we need both UNARY (Not) and N-ARY (NOT) operators--------
    __evalprint__(""" Not(AND([True,True])) == OR(NOT([True,True]))  """)
    __evalprint__(""" Not(AND([False,True])) == OR(NOT([False,True])) """)
    __evalprint__(""" Not(AND([True,False])) == OR(NOT([True,False])) """)
    __evalprint__(""" Not(AND([False,False])) == OR(NOT([False,False])) """)
    __evalprint__(""" TRUE == True """)
    __evalprint__(""" TRUE is True """)

## --------------------------------------------------
## --PLASM  Comparison operators---------------------
## --------------------------------------------------

def Eq (x,y): return x == y
def eQ (args): return args[0] == args[-1]

EQ = COMP([AND, AA(eQ), TRANS, CONS([RTAIL,TAIL])])

if __name__ == "__main__":
    print "\n## --PLASM  Comparison operators: EQ---------------------"
    __evalprint__(""" EQ([100, 10*10, 50+50, 200-100]) ### => True """)
    __evalprint__(""" EQ([100, 10*10, 50+50, 200-100]) ### => [100, 100, 100] !!!!! """)
    __evalprint__(""" EQ([100, 10*10])                 ### => [100] !!!! """)


NEQ = lambda x: COMP([NOT,LIST,EQ])(x)[0]

if __name__ == "__main__":
    print "\n## --PLASM  Comparison operators: NEQ---------------------"
    __evalprint__(""" NEQ([100, 10*10, 50+50, 200-100]) """)
    __evalprint__(""" NEQ([100, 10*10]) """)
           
def LT (a): 
    """
    LT: Standard core of PyPLaSM
    predicate that  tests if the second argument m is lower than n
    (n::isnum)(m::isnum) -> (isbool)
    """
    return lambda b: b < a

if __name__ == "__main__":
    print "\n## --PLASM  Comparison operators: LT---------------------"
    __evalprint__(""" LT(100)(200) == False """)
    __evalprint__(""" LT(200)(100) == True """)
    
def LE (a): 
    """
    Standard core of PyPLaSM
    less or equal than. Predicate that  tests if the second argument m is less or equal than n
    (n::isnum)(m::isnum) -> isbool
    """
    return lambda b: b <= a

if __name__ == "__main__":
    print "\n## --PLASM  Comparison operators: LE---------------------"
    __evalprint__(""" LE(100)(10*10) == True """)
    __evalprint__(""" LE(199)(100+100) == False """)
    
def GT (a): 
    """
    GT: Standard core of PyPLaSM
    greater than. Predicate testing if the second argument m is greater than n
    (n::isnum)(m::isnum) -> isbool
    """
    return lambda b: b > a
    
if __name__ == "__main__":
    print "\n## --PLASM  Comparison operators: GT---------------------"
    __evalprint__(""" GT(100)(200) == True """)
    __evalprint__(""" GT(200)(100) == False """)
    
def GE (a): 
    """
    GE: Standard core of PyPLaSM
    greater or equal. Predicate testing if the second argument m is greater or equal than n
    (n::isnum)(m::isnum) -> isbool
    """
    return lambda b: b >= a

if __name__ == "__main__":
    print "\n## --PLASM  Comparison operators: GE---------------------"
    __evalprint__(""" GE(100)(10*10) == True """)
    __evalprint__(""" GE(199)(100+100) == True """)

## --------------------------------------------------
## --PLASM  predicates-------------------------------
## --------------------------------------------------

def ISNUM (x):
    """
    ISNUM:  Standard core of PyPLaSM
    predicate that  tests if the argument a is a number
    (a::tt) -> (isbool)
    """
    return isinstance(x,int) or isinstance(x,long) or \
        isinstance(x,float) or isinstance(x,complex)

if __name__ == "__main__":
    print "\n## --PLASM  predicates: ISNUM-------------------------------"
    __evalprint__(""" ISNUM(10) """)
    __evalprint__(""" ISNUM(10000000000000000L) """)
    __evalprint__(""" ISNUM(PI) """)
    __evalprint__(""" ISNUM(10+0j) """)

def ISNUMPOS (x): 
    """
    ISNUMPOS:   Standard core of PyPLaSM
    predicate that  tests if the argument a} is a positive number
    (a::tt) -> (isbool)
    """
    return ISNUM(x) and x > 0

if __name__ == "__main__":
    print "\n## --PLASM  predicates: ISNUMPOS-------------------------------"
    __evalprint__(""" ISNUMPOS(PI) == True """)

def ISNUMNEG (x): 
    """
    ISNUMNEG:   Standard core of PyPLaSM
    predicate that  tests if the argument a} is a negative number
    (a::tt) -> (isbool)
    """
    return ISNUM(x) and x < 0

if __name__ == "__main__":
    print "\n## --PLASM  predicates: ISNUMNEG-------------------------------"
    __evalprint__(""" ISNUMNEG(PI) == False """)

def ISINT (x): 
    """
    ISINT:  Standard core of PyPLaSM
    predicate that  tests if the argument a} is an integer
    (a::tt) -> (isbool)
    """
    return isinstance(x,int)

if __name__ == "__main__":
    print "\n## --PLASM  predicates: ISINT-------------------------------"
    __evalprint__(""" ISINT(10) """)

def ISLONG (x): 
    """
    ISLONG:  Standard core of PyPLaSM
    predicate that  tests if the argument a is a long integer
    (a::tt) -> (isbool)
    """
    return isinstance(x,long)
    __evalprint__(""" ISLONG(10000000000000000L) """)

if __name__ == "__main__":
    print "\n## --PLASM  predicates: ISLONG-------------------------------"
    __evalprint__(""" ISLONG(10) == False """)

def ISINTPOS (x): 
    """
    ISINTPOS:   Standard core of PyPLaSM
    predicate that  tests if the argument a} is a positive integer
    (a::tt) -> (isbool)
    """
    return isinstance(x,int) and x > 0

if __name__ == "__main__":
    print "\n## --PLASM  predicates: ISINTPOS-------------------------------"
    __evalprint__(""" ISINTPOS(-10) """)

def ISINTNEG (x): 
    """
    ISINTNEG:   Standard core of PyPLaSM
    predicate that  tests if the argument a} is a negative integer
    (a::tt) -> (isbool)
    """
    return isinstance(x,int) and x < 0

if __name__ == "__main__":
    print "\n## --PLASM  predicates: ISINTNEG-------------------------------"

def ISREAL (x): 
    """
    ISREAL: Standard core of PyPLaSM
    predicate that  tests if the argument a} is a real number
    (a::tt) -> (isbool)
    """
    return isinstance(x,float)
    __evalprint__(""" ISREAL(PI) """)

if __name__ == "__main__":
    print "\n## --PLASM  predicates: ISREAL-------------------------------"

def ISREALPOS (x): 
    """
    ISREALPOS:  Standard core of PyPLaSM
    predicate that  tests if the argument a} is a positive real number
    (a::tt) -> (isbool)
    """
    return isinstance(x,float) and x > 0
    __evalprint__(""" ISREALPOS(-PI) """)

if __name__ == "__main__":
    print "\n## --PLASM  predicates: ISREALPOS-------------------------------"

def ISREALNEG (x): 
    """
    ISREALNEG:  Standard core of PyPLaSM
    predicate that  tests if the argument a} is a negative real number
    (a::tt) -> (isbool)
    """
    return isinstance(x,float) and x < 0

if __name__ == "__main__":
    print "\n## --PLASM  predicates: ISREALNEG-------------------------------"

def ISCOMPLEX (x): 
    """
    ISCOMPLEX:   Standard core of PyPLaSM
    predicate that  tests if the argument a is a complex number
    (a::tt) -> (isbool)
    """
    return isinstance(x,complex)

if __name__ == "__main__":
    print "\n## --PLASM  predicates: ISCOMPLEX-------------------------------"
    __evalprint__(""" ISCOMPLEX(10+0j) """)
        
def ISFUN (x):
    """
    ISFUN:  Standard core of PyPLaSM
    predicate that  tests if the argument a is a function
    (a::tt) -> (isbool)
    """
    return isinstance(x,types.LambdaType) or isinstance(x,types.FunctionType) 

if __name__ == "__main__":
    print "\n## --PLASM  predicates: ISFUN-------------------------------"
    __evalprint__(""" ISFUN(ID) """)
    __evalprint__(""" ISFUN(lambda x: x*x) """)
    __evalprint__(""" AA(ISFUN)([ID,K,I2,I1,AA,ISINT]) """)

def ISSEQ (x): 
    """
    ISSEQ:  Standard core of PyPLaSM
    predicate that  tests if the argument a is a sequence
    (a::tt) -> (isbool)
    """
    return isinstance(x,list)

if __name__ == "__main__":
    print "\n## --PLASM  predicates: ISSEQ-------------------------------"
    __evalprint__(""" ISSEQ(range(10)) """)

def ISSEQOF (type):
    """
    ISSEQOF:    Standard core of PyPLaSM
    second-order predicate that  tests if the argument a is a sequence with all elements of pred type
    (pred::isfun)(a::tt) -> isbool
    """
    return lambda Data: isinstance(Data,list) \
        and AND([type(e) for e in Data])

if __name__ == "__main__":
    __evalprint__(""" ISSEQOF(ISINT)(range(10)) """)
    __evalprint__(""" ISSEQOF(ISINT)(FROMTO([-5,5])) == True """)
    __evalprint__(""" ISSEQOF(ISINTPOS)(FROMTO([-5,5])) == False """)
    __evalprint__(""" ISSEQOF(ISREAL)([1.0,2.0,3.0]) == True """)
    print "\n## --PLASM  predicates: ISSEQOF-------------------------------"

def ISBOOL (x): 
    """
    ISBOOL: Standard core of PyPLaSM
    predicate that tests if the argument a is a boolean expression
    (a::tt) -> (isbool)
    """
    return isinstance(x,bool)

if __name__ == "__main__":
    print "\n## --PLASM  predicates: ISBOOL-------------------------------"
    __evalprint__(""" AA(ISBOOL)([TRUE,FALSE,TT,AND]) """)

def ISNULL (x): 
    """
    ISNULL: Standard core of PyPLaSM
    predicate that  tests if the argument a is the empy sequence
    (a::tt) -> (isbool)
    """
    return len(x) == 0

if __name__ == "__main__":
    print "\n## --PLASM  predicates: ISNULL-------------------------------"
    __evalprint__(""" AA(ISNULL)([[],(),{}]) """)

def ISPAIR (x): 
    """
    ISPAIR: Standard core of PyPLaSM
    predicate that  tests if the argument a} is a pair (a sequence of exacly two elements)
    (a::tt) -> (isbool)
    """
    return len(x) == 2

if __name__ == "__main__":
    print "\n## --PLASM  predicates: ISPAIR-------------------------------"
    __evalprint__(""" AA(ISPAIR)([[100,200],(100,200),{100:1,200:1},{100:1,200:2,300:3}]) """)

def ISCHAR (x): 
    """
    ISCHAR: Standard core of PyPLaSM
    predicate that  tests if the argument a is a character
    (arg::tt) -> (isbool)
    """
    return isinstance(x,str) and len(x) == 1

if __name__ == "__main__":
    print "\n## --PLASM  predicates: ISCHAR-------------------------------"
    __evalprint__(""" AA(ISCHAR)(['A','a',"0","?"]) """)

def ISSTRING (x): 
    """
    ISSTRING:   Standard core of PyPLaSM
    predicate that  tests if the argument a} is a string
    (a::tt) -> (isbool)
    """
    return isinstance(x,str)
    
if __name__ == "__main__":
    print "\n## --PLASM  predicates: ISSTRING-------------------------------"
    __evalprint__(""" ISSTRING("Alberto") """)


## --------------------------------------------------
## --Curryfing function------------------------------
## --------------------------------------------------

def C (fun):
    """
    C:  Standard core of PyPLaSM
    curryfies a function, so that, for example, fun([a,b]), can be evalued as c(fun)(a)(b)
    (fun::isfun) -> isfun
    """
    return lambda arg1: lambda arg2: fun([arg1,arg2])

if __name__ == "__main__":
    print "\n## --PLASM  binary currying-------------------------------"
    __evalprint__(""" C(EQ) """)
    __evalprint__(""" C(EQ)(0) """)
    __evalprint__(""" C(EQ)(0)(10-(2*5)) == True   """)
    __evalprint__(""" C(SUM)(1) """)
    __evalprint__(""" C(SUM)(1)(100) """)


## --------------------------------------------------
## --Miscellanea of 'standard' PLaSM functions-------
## --------------------------------------------------

def AS(fun):
    """
    AS: Standard core of PyPLaSM
    apply-in-sequence. AS:fun:seq is equivalent to(B ~ AA:fun:seq
    (fun::isfun)(seq::isseq) -> (isfun)
    """
    return lambda args: COMP([CONS, AA(fun)])(args)

if __name__ == "__main__":
    print "\n## --Miscellanea of 'standard' functions: AS-------"
    __evalprint__(""" AS(SEL)(range(4,16,4)) """)
    __evalprint__(""" AS(SEL)(range(4,16,4))(range(-10,10)) """)

def AC(fun):
    """
    AC: Standard core of PyPLaSM
    apply-in-composition. AC:fun:seq is equivalent to (COMP ~ AA:fun):seq
    (fun::isfun)(seq::isseq) -> (isfun)
    """
    return lambda args: COMP(AA(fun)(args))

if __name__ == "__main__":
    print "\n## --Miscellanea of 'standard' functions: AC-------"
    __evalprint__(""" AC(SEL) """)
    __evalprint__(""" AC(SEL)([3,1]) """)
    __evalprint__(""" AC(SEL)([3,1])([[1,3,8,7],89,"fun"]) """)

def CHARSEQ (String):
    """
    CHARSEQ:    Standard core of PyPLaSM
    maps a string into a sequence of characters
    (str::isstring) -> (isseqof:ischar)
    """
    return [String[i] for i in range(len(String))]

if __name__ == "__main__":
    print "\n## --Miscellanea of 'standard' functions: CHARSEQ-------"
    __evalprint__(""" CHARSEQ("Ciao") """)
    __evalprint__(""" CHARSEQ("Alberto: 'sono' io") """)
    __evalprint__(""" CHARSEQ('Alberto: "sono" io') """)

def STRING (Charseq): 
    """
    STRING: Standard core of PyPLaSM
    maps a sequence of characters into a string
    (chars::isseqof:ischar) -> (isstring)
    """
    return SUM(Charseq)

if __name__ == "__main__":
    print "\n## --Miscellanea of 'standard' functions: STRING-------"
    __evalprint__(""" STRING(CHARSEQ('Alberto: "sono" io')) """)

def RANGE (Pair):
    """
    RANGE:  Standard core of PyPLaSM
    returns the integer sequence (possibly reversed) from m} to n}
    (m,n::isint) -> (isseq)
    """
    if ( Pair[-1]-Pair[0] >= 0 ): return range(Pair[0], Pair[-1] + 1)
    elif ( Pair[-1]-Pair[0] <= -1 ): return range(Pair[0], Pair[-1] - 1, -1)

if __name__ == "__main__":
    print "\n## --Miscellanea of 'standard' functions: RANGE-------"
    __evalprint__(""" RANGE([-5,5]) """)
    __evalprint__(""" RANGE([5,-5]) """)
    __evalprint__(""" RANGE([5,5]) """)
    __evalprint__(""" RANGE([0,0]) """)

def SIGN (Number): 
    """
    SIGN:   Standard core of PyPLaSM
    returns either 1 if x is positive, or -1 if x is negative, or 0 if x is zero
    (x::isnum) -> isint
    """
    return int(Number/abs(Number))

if __name__ == "__main__":
    print "\n## --Miscellanea of 'standard' functions: SIGN-------"
    __evalprint__(""" SIGN(-10) """)
    __evalprint__(""" SIGN(10) """)
    __evalprint__(""" SIGN(-10.123) """)
    __evalprint__(""" SIGN(10.123) """)

def PRINT (AnyValue):
    """
    PRINT:  Standard core of PyPLaSM
    returns arg} and prints its value in the listener. It may be used for debugging
    (arg::tt) -> (tt)
    """
    print(AnyValue)
    return AnyValue

if __name__ == "__main__":
    print "\n## --Miscellanea of 'standard' functions: PRINT-------"
    __evalprint__(""" COMP([SIN,PRINT,COS,PRINT,TAN,PRINT,ID])(1.0) """)
    __evalprint__(""" COMP([TAIL,PRINT,RTAIL,PRINT,S1,PRINT,N(2),PRINT,ID])(range(10)) """)


## --------------------------------------------------
## --IF THEN ELSE purely functional------------------
## --------------------------------------------------

def IF (pred):
    """
    IF:  Standard core of PyPLaSM
    conditional function. It is applied to a triple of functions, finally applied to the x argument
    (pred::isfun, then::isfun, else::isfun)(x::tt) -> (tt)
    """
    return lambda fpair: lambda x: (pred(x) and fpair[0](x)) \
            or (not(pred(x)) and fpair[-1](x))

def IF (Funs):
    def IF_NO_CURRIED (Funs,x):
        f1, f2, f3 = Funs
        return f2(x) if f1(x) else f3(x)
    return lambda x: IF_NO_CURRIED(Funs,x)

        
if __name__ == "__main__":
    print "\n## --IF THEN ELSE short circuit: Lambda style-------"
    __evalprint__(""" IF([GT(0), SQRT, K(0)])(9) """)
    __evalprint__(""" IF([GT(0), SQRT, K(0)])(-99) """)
    
    

# def CASE (Pairs):
#     """
#     CASE:   Standard core of PyPLaSM
#     arguments are pairs (pred_i, fun_i)} to be tested in sequence, until the first is true. 
#     (conds::isseqof:(ispred ~ s1, isfun ~ s2)(x::t) -> (isfun)
#     """
#     return lambda x: (Funs[0](x) and Funs[1](x)) \
#            or (not(Funs[0](x)) and Funs[2](x))


## --------------------------------------------------
## --FL LIFT and RAISE functions---------------------
## --------------------------------------------------




def LIFT (f):
    """
    LIFT:   Standard core of PyPLaSM
    the semantics of this combining form is: 
    LIFT(SUM)(funs) == COMP([f, CONS(funs)])
    (f::isfun)(funs::isseqof:isfun) -> isfun
    """
    return lambda funs: COMP([f, CONS(funs)])

def LIFT (f):
    return lambda funs: COMP([f, CONS(funs)])

if __name__ == "__main__":
    print "\n## --FL LIFT function-------------------------------"
    __evalprint__(""" LIFT """)
    __evalprint__(""" LIFT(SUM)([SIN,COS]) """)
    __evalprint__(""" LIFT(SUM)([SIN,COS])(0.0) - LIFT(SUM)([SIN,COS])(PI/2) """)
    __evalprint__(""" LIFT(SUM)([SIN,COS])(PI/4) """)
    
def RAISE (f):
    """
    RAISE:   Standard core of PyPLaSM
    combinating form  used for allowing overloaded use of operators over
    both numbers and functions. 
    (f::isfun)(args::isseq) -> (isfun)
    """
    return lambda args: IF(ISSEQOF(ISFUN))([LIFT(f),f])(args)


def RAISE (f):
    return lambda args: IF([ISSEQOF(ISFUN), LIFT(f), f])(args)

if __name__ == "__main__":
    print "\n## --FL  RAISE function------------------------------"
    __evalprint__(""" RAISE """)
    __evalprint__(""" RAISE(SUM) """)
    __evalprint__(""" RAISE(SUM)([SUM,PROD]) """)
    __evalprint__(""" RAISE(SUM)([SUM,PROD])([3,5]) """)
    __evalprint__(""" RAISE(SUM)([SUM,PROD])([3,5]) == COMP([SUM, CONS([SUM,PROD])])([3,5]) """)


def SQR (x):
    """ Raised squaring operator.

    return either a function or a number depending on type(x)
    """
    return RAISE(PROD)([ID,ID])(x)

if __name__ == "__main__":
    print "\n## --FL  SQR == ID * ID------------------------------"
    __evalprint__(""" SQR(4) """)
    __evalprint__(""" SQR(10) """)
##    __evalprint__(""" SQR(K) """)
##    __evalprint__(""" SQR(K)(4) """)


def INNERPROD (args):
    return COMP([RAISE(SUM),AA(RAISE(PROD)),TRANS])(args)

if __name__ == "__main__":
    print "\n## --FL INNERPROD (on vectors of numbers or functions)--"
    __evalprint__(""" INNERPROD([[1,2,3,4,5],[1,2,3,4,5]]) """)
    __evalprint__(""" RAISE(INNERPROD)([CONS([SIN,COS]), CONS([SIN,COS])]) """)
    __evalprint__(""" RAISE(INNERPROD)([CONS([SIN,COS]), CONS([SIN,COS])])(PI) """)


def FROMTO (args):
    """
    FROMTO: Standard core of PyPLaSM
    returns the integer sequence from m to n. 
    When m < n  the returned sequence is empty.
    (m,n::isint) -> isseqof:isint
    """
    n,m = args
    return range(n,m+1)
